/*
    check.c
	various functions dealing with checking time.
    
    Copyright 2003 - 2004 Night Productions, by Michael A. Nickerson (aka Darkshadow).  All rights reserved.
    Licensed under the Open Software License version 2.0

    Teaching a program how to tell time is one giant pain in the ass.
    I had to literally forget how to tell time then teach myself again
    in terms that the computer would understand before I could get it.
    Definitely not easy.
    
    This program uses a two-tier approach:  A basic test to tell
    if the time has passed, then a more defined test to insure that
    the time is within the boundries.
*/


#include <ctype.h>
#include "loop.h"

#define HourDef			3600L
#define DayDef			86400L
#define WeekDef			604800L
#define MonthDef		2678400L  /* 31 Days */
#define YearDef			31536000L /* 365 Days */
#define CheckStarDef(s)		(*(s) == '*')
#define ListSpanRangeTestDef(s) (strchr(s,',') != NULL) || (strchr(s,'-') != NULL) || (strchr(s,'/') != NULL)
#define CheckEmptyDef(s)	((*(s) == '*') && !(ListSpanRangeTestDef((s))))

static int lastMin = -1, lastHour = -1, lastMonth = 0, lastMDay = -1, lastWDay = -1;

int returnLists(char *digits, char *words[], int max) {
    char *p = digits;
    int i = 0;
    
    while (1) {
	while((*p == ',')) {
	    p++;
	}
	if (*p == '\0') {
	    return i;
	}
	words[i++] = p;
	while(!(*p == ',') && *p != '\0') {
	    p++;
	}
	if (*p == '\0') {
	    return i;
	}
	if (i >= max) {
	    return i;
	}
	*p++ = '\0';
    }
}

int returnSpans(char *digits, char *words[], int max) {
    char *p = digits;
    int i = 0;
    
    while (1) {
	while(((*p == '-') || (*p == '/'))) {
	    p++;
	}
	if (*p == '\0') {
	    return i;
	}
	words[i++] = p;
	while(!((*p == '-') || (*p == '/')) && *p != '\0') {
	    p++;
	}
	if (*p == '\0') {
	    return i;
	}
	if (i >= max) {
	    return i;
	}
	*p++ = '\0';
    }
}
    
int checkDayOrMonth(char *name)
{
    if (!strcmp(name, "Sun") || !strcmp(name, "sun") || !strcmp(name, "SUN")) {
	return 0;
    } else if (!strcmp(name, "Mon") || !strcmp(name, "mon") || !strcmp(name, "MON") || !strcmp(name, "Jan") || !strcmp(name, "jan") || !strcmp(name, "JAN")) {
	return 1;
    } else if (!strcmp(name, "Tue") || !strcmp(name, "tue") || !strcmp(name, "TUE") || !strcmp(name, "Feb") || !strcmp(name, "feb") || !strcmp(name, "FEB")) {
	return 2;
    } else if (!strcmp(name, "Wed") || !strcmp(name, "wed") || !strcmp(name, "WED") || !strcmp(name, "Mar") || !strcmp(name, "mar") || !strcmp(name, "MAR")) {
	return 3;
    } else if (!strcmp(name, "Thu") || !strcmp(name, "thu") || !strcmp(name, "THU") || !strcmp(name, "Apr") || !strcmp(name, "apr") || !strcmp(name, "APR")) {
	return 4;
    } else if (!strcmp(name, "Fri") || !strcmp(name, "fri") || !strcmp(name, "FRI") || !strcmp(name, "May") || !strcmp(name, "may") || !strcmp(name, "MAY")) {
	return 5;
    } else if (!strcmp(name, "Sat") || !strcmp(name, "sat") || !strcmp(name, "SAT") || !strcmp(name, "Jun") || !strcmp(name, "jun") || !strcmp(name, "JUN")) {
	return 6;
    } else if (!strcmp(name, "Jul") || !strcmp(name, "jul") || !strcmp(name, "JUL")) {
	return 7;
    } else if (!strcmp(name, "Aug") || !strcmp(name, "aug") || !strcmp(name, "AUG")) {
	return 8;
    } else if (!strcmp(name,"Sep") || !strcmp(name,"sep") || !strcmp(name,"SEP")) {
	return 9;
    } else if (!strcmp(name,"Oct") || !strcmp(name,"oct") || !strcmp(name,"OCT")) {
	return 10;
    } else if (!strcmp(name,"Nov") || !strcmp(name,"nov") || !strcmp(name,"NOV")) {
	return 11;
    } else if (!strcmp(name,"Dec") || !strcmp(name,"dec") || !strcmp(name,"DEC")) {
	return 12;
    }
    return -1000;
}

int checkPass(int testNum, struct tm *ltm, struct tm *ntm, int topNum, long secs)
{
    int nowNum = -1, lastNum = -1;
#ifdef USE_LOG
        FILE *log = fopen(loopLogDef,"a");
#endif

    switch (topNum) {
	case 59:
	    nowNum = ntm->tm_min;
	    lastNum = ltm->tm_min;
	    if (secs >= HourDef) {
		/* If it's been longer than an hour, then the minutes
		    pass. */
#ifdef USE_LOG
		fprintf(log, "Longer than an hour, minutes pass.\n");
		fclose(log);
#endif
		return 1;
	    } else if (ltm->tm_hour != ntm->tm_hour) {
		/* It's been less than an hour, but the hour has changed.
		   For a goodly while, I couldn't figure out how to get this.
		   Then this solution finally hit me.  It works.  It's good.
		   That's all I needed. */
		int dif, testDif, nowDif;
		dif = (60 - ltm->tm_min);
		testDif = (testNum + dif);
		nowDif = (ntm->tm_min + dif);
		if (testDif >= 60) {
		    testDif = (testDif - 60);
		}
		if (nowDif >= 60) {
		    nowDif = (nowDif - 60);
		}
#ifdef USE_LOG
		fprintf(log, "\tlastNum: %d\tdif: %d\n\ttestNum: %d\ttestDif: %d\n\tnowNum: %d\tnowDif: %d\n", ltm->tm_min, dif, testNum, testDif, ntm->tm_min, nowDif);
		fprintf(log, "(%d <= %d) [%d]\n", testDif, nowDif, (testDif <= nowDif));
		fclose(log);
#endif
		return (testDif <= nowDif);
	    }
	    break;
	case 23:
	    nowNum = ntm->tm_hour;
	    lastNum = ltm->tm_hour;
	    if (secs >= DayDef) {
		/* If it's been over a day, then the hours pass. */
#ifdef USE_LOG
		fprintf(log, "Longer than a day, hours pass.\n");
		fclose(log);
#endif
		return 1;
	    } else if (ltm->tm_yday != ntm->tm_yday) {
		int dif, testDif, nowDif;
		dif = (24 - ltm->tm_hour);
		testDif = (testNum + dif);
		nowDif = (ntm->tm_hour + dif);
		if (testDif >= 24) {
		    testDif = (testDif - 24);
		}
		if (nowDif >= 24) {
		    nowDif = (nowDif - 24);
		}
#ifdef USE_LOG
		fprintf(log, "\tlastNum: %d\tdif: %d\n\ttestNum: %d\ttestDif: %d\n\tnowNum: %d\tnowDif: %d\n", ltm->tm_hour, dif, testNum, testDif, ntm->tm_hour, nowDif);
		fprintf(log, "(%d <= %d) [%d]\n", testDif, nowDif, (testDif <= nowDif));
		fclose(log);
#endif
		return (testDif <= nowDif);
	    }
	    break;
	case 31:
	    nowNum = ntm->tm_mday;
	    lastNum = ltm->tm_mday;
	    if (secs >= MonthDef) {
		/* If it's been over a month, then the month days pass as well.
		    Note that I'm generalizing here, using 31 days as the test
		    when not every month has 31 days...but then again, who leaves their
		    machine asleep/shut down for that long? */
#ifdef USE_LOG
		fprintf(log, "Longer than a month, any month day passes.\n");
		fclose(log);
#endif
		return 1;
	    } else if (ltm->tm_mon != ntm->tm_mon) {
		int dif, testDif, nowDif;
		dif = (31 - ltm->tm_mday);
		testDif = (testNum + dif);
		nowDif = (ntm->tm_mday + dif);
		if (testDif >= 31) {
		    testDif = (testDif - 31);
		}
		if (nowDif >= 31) {
		    nowDif = (nowDif - 31);
		}
#ifdef USE_LOG
		fprintf(log, "\tlastNum: %d\tdif: %d\n\ttestNum: %d\ttestDif: %d\n\tnowNum: %d\tnowDif: %d\n", ltm->tm_mday, dif, testNum, testDif, ntm->tm_mday, nowDif);
		fprintf(log, "(%d <= %d) [%d]\n", testDif, nowDif, (testDif <= nowDif));
		fclose(log);
#endif
		return (testDif <= nowDif);
	    }
	    break;
	case 12:
	    /* Note that I'm not testing the seconds here - it would be a test
		for over a year for the months, and I've already done that.
		This block of code will only be reached if the year has changed but
		it hasn't actually been a year since the program last ran. */
	    nowNum = ntm->tm_mon;
	    lastNum = ltm->tm_mon;
	    testNum = testNum - 1;
	    if (ltm->tm_year != ntm->tm_year) {
		int dif, testDif, nowDif;
		dif = (11 - ltm->tm_mon);
		testDif = (testNum + dif);
		nowDif = (ntm->tm_mon + dif);
		if(testDif >= 11) {
		    testDif = (testDif - 11);
		}
		if (nowDif >= 11) {
		    nowDif = (nowDif - 11);
		}
#ifdef USE_LOG
		fprintf(log, "\tlastNum: %d\tdif: %d\n\ttestNum: %d\ttestDif: %d\n\tnowNum: %d\tnowDif: %d\n", ltm->tm_mon, dif, testNum, testDif, ntm->tm_mon, nowDif);
		fprintf(log, "(%d <= %d) [%d]\n", testDif, nowDif, (testDif <= nowDif));
		fclose(log);
#endif
		return (testDif <= nowDif);
	    }
	case 7:
	    nowNum = ntm->tm_wday;
	    lastNum = ltm->tm_wday;
	    if (secs >= WeekDef) {
		/* Week days pass if it's been over a week. */
#ifdef USE_LOG
		fprintf(log, "Longer than a week, weekdays pass.\n");
		fclose(log);
#endif
		return 1;
	    } else if (ltm->tm_wday > ntm->tm_wday) {
		int dif, testDif, nowDif;
		dif = (7 - ltm->tm_wday);
		testDif = (testNum + dif);
		nowDif = (ntm->tm_wday + dif);
		if(testDif >= 7) {
		    testDif = (testDif - 7);
		}
		if (nowDif >= 7) {
		    nowDif = (nowDif - 7);
		}
#ifdef USE_LOG
		fprintf(log, "\tlastNum: %d\tdif: %d\n\ttestNum: %d\ttestDif: %d\n\tnowNum: %d\tnowDif: %d\n", ltm->tm_wday, dif, testNum, testDif, ntm->tm_wday, nowDif);
		fprintf(log, "(%d <= %d) [%d]\n", testDif, nowDif, (testDif <= nowDif));
		fclose(log);
#endif
		return (testDif <= nowDif);
	    }
	    break;
	default:
	    break;
    }
    /* Test to see if either now is the time it should be
       run, or test the range between the last time and now
       to see if it should run. */

#ifdef USE_LOG
	fprintf(log, "(%d == %d) || ((%d >= %d) && (%d <= %d)) [%d]\n", testNum, nowNum, testNum, lastNum, testNum, nowNum, ((testNum >= lastNum) && (testNum <= nowNum)));
	fclose(log);
#endif
	return ((testNum == nowNum) || ((testNum >= lastNum) && (testNum <= nowNum)));

}

int checkSpan(char *inSpan, struct tm *ltm, struct tm *ntm, int top, long secs)
{
    int i = 0, j = 0, k = 0, check = 0;
    int range = (strchr(inSpan, '-') != NULL);
    int span = (strchr(inSpan, '/') != NULL);
    int list = (strchr(inSpan, ',') != NULL);
    int rangeFirst, rangeLast, spanDig, diff;
    char *words[121];
    char minutes[121];
    
    strcpy(minutes,inSpan);
    
    if (list) {
	i = returnLists(minutes, words, 121);
	for (j = 0; j < i; j++) {
	    if (span && range) {
		if ((strchr(words[j], '-') != NULL) && (strchr(words[j], '/') != NULL)) {
		    char *swords[10];
		    returnSpans(words[j], swords,10);
		    rangeFirst = strtol(swords[0], NULL, 10);
		    rangeLast = strtol(swords[1], NULL, 10);
		    spanDig = strtol(swords[2], NULL, 10);
		    
		    /*is it even possible to set a span/range on week/month >>Names<< ?*/
		    /*Answer to self: No.  But since I already implemented this, I guess
		      I'll leave it.  Maybe it'll be changed sometime and then I won't have
		      to worry about it. :p */
		    if (isalpha(*swords[0])) {
			rangeFirst = checkDayOrMonth(swords[0]);
		    }
		    if (isalpha(*swords[1])) {
			rangeLast = checkDayOrMonth(swords[1]);
		    }
		    diff = (0 - rangeFirst);
		    rangeFirst = (rangeFirst + diff);
		    rangeLast = (rangeLast + diff);
		    while((rangeFirst <= rangeLast)) {
		        if ((rangeFirst % spanDig) == 0) {
			    check = checkPass((rangeFirst - diff), ltm, ntm, top, secs);
			    if (check) {
				int saved = rangeFirst;
				rangeFirst++;
				while((rangeFirst <= rangeLast)) {
				    if ((rangeFirst % spanDig) == 0) {
					check = checkPass((rangeFirst - diff), ltm, ntm, top, secs);
					if (check) {
					    switch(top) {
						case 59:
						    lastMin = (saved - diff);
						    break;
						case 23:
						    lastHour = (saved - diff);
						    break;
						case 31:
						    lastMDay = (saved - diff);
						    break;
						case 12:
						    lastMonth = (saved - diff);
						    break;
						case 7:
						    lastWDay = (saved - diff);
						    break;
					    }
					    return (rangeFirst - diff);
					}
					break;
				    }
				}
				return (saved - diff);
			    }
			}
	                rangeFirst++;
	            }
		} else if ((strchr(words[j],'-') != NULL)) {
		    char *swords[10];
		    returnSpans(words[j],swords,10);
		    rangeFirst = strtol(swords[0],NULL,10);
		    rangeLast = strtol(swords[1],NULL,10);
		    
		    if (isalpha(*swords[0])) {
			rangeFirst = checkDayOrMonth(swords[0]);
		    }
		    if (isalpha(*swords[1])) {
			rangeFirst = checkDayOrMonth(swords[1]);
		    }
		    
		    while((rangeFirst <= rangeLast)) {
			check=checkPass(rangeFirst, ltm, ntm, top, secs);
			if (check) {
			    int saved = rangeFirst;
			    rangeFirst++;
			    check = checkPass(rangeFirst, ltm, ntm, top, secs);
			    if (check) {
				switch(top) {
				    case 59:
					lastMin = saved;
					break;
				    case 23:
					lastHour = saved;
					break;
				    case 31:
					lastMDay = saved;
					break;
				    case 12:
					lastMonth = saved;
					break;
				    case 7:
					lastWDay = saved;
					break;
				}
				return rangeFirst;
			    }
			    return saved;
			}
			rangeFirst++;
		    }
		} else if ((strchr(words[j], '/') != NULL) && (strchr(words[j], '*') != NULL)) {
		    char *swords[10];
		    returnSpans(words[j], swords, 10);
		    rangeFirst = 0;
		    rangeLast = (top == 12) ? 11 : top;
		    spanDig = strtol(swords[1], NULL, 10);
		    diff = (top == 12) ? 1 : 0;
		    
		    while ((rangeFirst <= rangeLast)) {
			if ((rangeFirst % spanDig) == 0) {
			    check = checkPass((rangeFirst + diff), ltm, ntm, top, secs);
			    if (check) {
				int saved = rangeFirst;
				rangeFirst++;
				while ((rangeFirst <= rangeLast)) {
				    if ((rangeFirst % spanDig) == 0) {
					check = checkPass((rangeFirst + diff), ltm, ntm, top, secs);
					if (check) {
					    switch(top) {
						case 59:
						    lastMin = (saved + diff);
						    break;
						case 23:
						    lastHour = (saved + diff);
						    break;
						case 31:
						    lastMDay = (saved + diff);
						    break;
						case 12:
						    lastMonth = (saved + diff);
						    break;
						case 7:
						    lastWDay = (saved + diff);
						    break;
					    }
					    return (rangeFirst + diff);
					}
					break;
				    }
				    rangeFirst++;
				}
				return (saved + diff);
			    }
			}
			rangeFirst++;
		    }
		} else if ((strchr(words[j], '/') != NULL)) {
		    char *swords[10];
		    returnSpans(words[j], swords, 10);
		    k = 0;
		    rangeFirst = strtol(swords[0], NULL, 10);
		    spanDig = strtol(swords[1], NULL, 10);
		    diff = (top == 12) ? 1 : 0;
		    
		    if (isalpha(*swords[0])) {
			rangeFirst = checkDayOrMonth(swords[0]);
		    }
		    if (diff) { rangeFirst--; }
		    
		    while((k <= rangeFirst)) {
			if ((k % spanDig) == 0) {
			    check = checkPass((k + diff), ltm, ntm, top, secs);
			    if (check) {
				int saved = k;
				k++;
				while ((k <= rangeFirst)) {
				    if ((k % spanDig) == 0) {
					check = checkPass((k + diff), ltm, ntm, top, secs);
					if (check) {
					    switch(top) {
						case 59:
						    lastMin = (saved + diff);
						    break;
						case 23:
						    lastHour = (saved + diff);
						    break;
						case 31:
						    lastMDay = (saved + diff);
						    break;
						case 12:
						    lastMonth = (saved + diff);
						    break;
						case 7:
						    lastWDay = (saved + diff);
						    break;
					    }
					    return (k + diff);
					}
					break;
				    }
				    k++;
				}
				return (saved + diff);
			    }
			}
			k++;
		    }
		} else {
		    rangeFirst = strtol(words[j], NULL, 10);
		    
		    if (isalpha(*words[j])) {
			rangeFirst = checkDayOrMonth(words[j]);
		    }
		    
		    check = checkPass(rangeFirst, ltm, ntm, top, secs);
		    if (check) {
			return rangeFirst;
		    }
		}
	    } else if (span) {
		if ((strchr(words[j], '/') != NULL) && (strchr(words[j], '*') != NULL)) {
		    char *swords[10];
		    returnSpans(words[j], swords, 10);
		    rangeFirst = 0;
		    rangeLast = (top == 12) ? 11 : top;
		    spanDig = strtol(swords[1], NULL, 10);
		    diff = (top == 12) ? 1 : 0;
		    
		    while ((rangeFirst <= rangeLast)) {
			if ((rangeFirst % spanDig) == 0) {
			    check = checkPass((rangeFirst + diff), ltm, ntm, top, secs);
			    if (check) {
				int saved = rangeFirst;
				rangeFirst++;
				while ((rangeFirst <= rangeLast)) {
				    if ((rangeFirst % spanDig) == 0) {
					check = checkPass((rangeFirst + diff), ltm, ntm, top, secs);
					if (check) {
					    switch(top) {
						case 59:
						    lastMin = (saved + diff);
						    break;
						case 23:
						    lastHour = (saved + diff);
						    break;
						case 31:
						    lastMDay = (saved + diff);
						    break;
						case 12:
						    lastMonth = (saved + diff);
						    break;
						case 7:
						    lastWDay = (saved + diff);
						    break;
					    }
					    return (rangeFirst + diff);
					}
					break;
				    }
				    rangeFirst++;
				}
				return (saved + diff);
			    }
			}
			rangeFirst++;
		    }
		} else if ((strchr(words[j], '/') != NULL)) {
		    char *swords[10];
		    returnSpans(words[j], swords, 10);
		    k = 0;
		    rangeFirst = strtol(swords[0], NULL, 10);
		    spanDig = strtol(swords[1], NULL, 10);
		    diff = (top == 12) ? 1 : 0;
		    
		    if (isalpha(*swords[0])) {
			rangeFirst = checkDayOrMonth(swords[0]);
		    }
		    if (diff) { rangeFirst--; }
		    
		    while((k <= rangeFirst)) {
			if ((k % spanDig) == 0) {
			    check = checkPass((k + diff), ltm, ntm, top, secs);
			    if (check) {
				int saved = k;
				k++;
				while ((k <= rangeFirst)) {
				    if ((k % spanDig) == 0) {
					check = checkPass((k + diff), ltm, ntm, top, secs);
					if (check) {
					    switch(top) {
						case 59:
						    lastMin = (saved + diff);
						    break;
						case 23:
						    lastHour = (saved + diff);
						    break;
						case 31:
						    lastMDay = (saved + diff);
						    break;
						case 12:
						    lastMonth = (saved + diff);
						    break;
						case 7:
						    lastWDay = (saved + diff);
						    break;
					    }
					    return (k + diff);
					}
					break;
				    }
				    k++;
				}
				return (saved + diff);
			    }
			}
			k++;
		    }
		} else {
		    rangeFirst = strtol(words[j], NULL, 10);
		    
		    if (isalpha(*words[j])) {
			rangeFirst = checkDayOrMonth(words[j]);
		    }
		    
		    check = checkPass(rangeFirst, ltm, ntm, top, secs);
		    if (check) {
			return rangeFirst;
		    }
		}
	    } else if (range) {
		if ((strchr(words[j],'-') != NULL)) {
		    char *swords[10];
		    returnSpans(words[j], swords, 10);
		    rangeFirst = strtol(swords[0], NULL, 10);
		    rangeLast = strtol(swords[1], NULL, 10);
		    
		    if (isalpha(*swords[0])) {
			rangeFirst = checkDayOrMonth(swords[0]);
		    }
		    if (isalpha(*swords[1])) {
			rangeLast = checkDayOrMonth(swords[1]);
		    }
		    
		    while((rangeFirst <= rangeLast)) {
			check = checkPass(rangeFirst, ltm, ntm, top, secs);
			if (check) {
			    int saved = rangeFirst;
			    rangeFirst++;
			    check = checkPass(rangeFirst, ltm, ntm, top, secs);
			    if (check) {
				switch(top) {
				    case 59:
					lastMin = saved;
					break;
				    case 23:
					lastHour = saved;
					break;
				    case 31:
					lastMDay = saved;
					break;
				    case 12:
					lastMonth = saved;
					break;
				    case 7:
					lastWDay = saved;
					break;
				}
				return rangeFirst;
			    }
			    return saved;
			}
			rangeFirst++;
		    }
		} else {
		    rangeFirst = strtol(words[j], NULL, 10);
		    
		    if (isalpha(*words[j])) {
			rangeFirst = checkDayOrMonth(words[j]);
		    }
		    
		    check = checkPass(rangeFirst, ltm, ntm, top, secs);
		    if (check) {
			return rangeFirst;
		    }
		}
	    } else {
		rangeFirst = strtol(words[j], NULL, 10);
		
		if (isalpha(*words[j])) {
		    rangeFirst = checkDayOrMonth(words[j]);
		}
		
		check = checkPass(rangeFirst, ltm, ntm, top, secs);
		if (check) {
		    return rangeFirst;
		}
	    }
	}
    } else if (range) {
	if ((strchr(minutes, '/') != NULL)) {
	    returnSpans(minutes, words, 121);
	    rangeFirst = strtol(words[0], NULL, 10);
	    rangeLast = strtol(words[1], NULL, 10);
	    spanDig = strtol(words[2], NULL, 10);
	    
	    if (isalpha(*words[0])) {
		rangeFirst = checkDayOrMonth(words[0]);
	    }
	    if (isalpha(*words[1])) {
		rangeFirst = checkDayOrMonth(words[1]);
	    }
	    
	    diff = (0 - rangeFirst);
	    rangeFirst = (rangeFirst + diff);
	    rangeLast = (rangeLast + diff);
	    
	    while ((rangeFirst <= rangeLast)) {
		if ((rangeFirst % spanDig) == 0) {
		    check = checkPass((rangeFirst - diff), ltm, ntm, top, secs);
		    if (check) {
			int saved = rangeFirst;
			rangeFirst++;
			while ((rangeFirst <= rangeLast)) {
			    if ((rangeFirst % spanDig) == 0) {
				check = checkPass((rangeFirst - diff), ltm, ntm, top, secs);
				if (check) {
				    switch(top) {
					case 59:
					    lastMin = (saved - diff);
					    break;
					case 23:
					    lastHour = (saved - diff);
					    break;
					case 31:
					    lastMDay = (saved - diff);
					    break;
					case 12:
					    lastMonth = (saved - diff);
					    break;
					case 7:
					    lastWDay = (saved - diff);
					    break;
				    }
				    return (rangeFirst - diff);
				}
				break;
			    }
			    rangeFirst++;
			}
			return (saved - diff);
		    }
		}
		rangeFirst++;
	    }
	} else {
	    returnSpans(minutes, words, 121);
	    rangeFirst = strtol(words[0], NULL, 10);
	    rangeLast = strtol(words[1], NULL, 10);
	    
	    if (isalpha(*words[0])) {
		rangeFirst = checkDayOrMonth(words[0]);
	    }
	    if (isalpha(*words[1])) {
		rangeLast = checkDayOrMonth(words[1]);
	    }
	    
	    while ((rangeFirst <= rangeLast)) {
		check = checkPass(rangeFirst, ltm, ntm, top, secs);
		if (check) {
		    int saved = rangeFirst;
		    rangeFirst++;
		    check = checkPass(rangeFirst, ltm, ntm, top, secs);
		    if (check) {
			switch(top) {
			    case 59:
				lastMin = saved;
				break;
			    case 23:
				lastHour = saved;
				break;
			    case 31:
				lastMDay = saved;
				break;
			    case 12:
				lastMonth = saved;
				break;
			    case 7:
				lastWDay = saved;
				break;
			}
			return rangeFirst;
		    }
		    return saved;
		}
		rangeFirst++;
	    }
	}
    } else if (span) {
	if (strchr(minutes,'*') != NULL) {
	    returnSpans(minutes, words, 5);
	    rangeFirst = 0;
	    rangeLast = (top == 12) ? 11 : top;
	    spanDig = strtol(words[1], NULL, 10);
	    diff = (top == 12) ? 1 : 0;
	    
	    while ((rangeFirst <= rangeLast)) {
		if ((rangeFirst % spanDig) == 0) {
		    check = checkPass((rangeFirst + diff), ltm, ntm, top, secs);
		    if (check) {
			int saved = rangeFirst;
			rangeFirst++;
			while ((rangeFirst <= rangeLast)) {
			    if ((rangeFirst % spanDig) == 0) {
				check = checkPass((rangeFirst + diff), ltm, ntm, top, secs);
				if (check) {
				    switch(top) {
					case 59:
					    lastMin = (saved + diff);
					    break;
					case 23:
					    lastHour = (saved + diff);
					    break;
					case 31:
					    lastMDay = (saved + diff);
					    break;
					case 12:
					    lastMonth = (saved + diff);
					    break;
					case 7:
					    lastWDay = (saved + diff);
					    break;
				    }
				    return (rangeFirst + diff);
				}
				break;
			    }
			    rangeFirst++;
			}
			return (saved + diff);
		    }
		}
		rangeFirst++;
	    }
	} else if ((strchr(minutes, '-') != NULL)) {
	    returnSpans(minutes, words, 121);
	    rangeFirst = strtol(words[0], NULL, 10);
	    rangeLast = strtol(words[1], NULL, 10);
	    spanDig = strtol(words[2], NULL, 10);
	    
	    if (isalpha(*words[0])) {
		rangeFirst = checkDayOrMonth(words[0]);
	    }
	    if (isalpha(*words[1])) {
		rangeFirst = checkDayOrMonth(words[1]);
	    }
	    
	    diff = (0 - rangeFirst);
	    rangeFirst = (rangeFirst + diff);
	    rangeLast = (rangeLast + diff);
	    
	    while ((rangeFirst <= rangeLast)) {
		if ((rangeFirst % spanDig) == 0) {
		    check = checkPass((rangeFirst - diff), ltm, ntm, top, secs);
		    if (check) {
			int saved = rangeFirst;
			rangeFirst++;
			while ((rangeFirst <= rangeLast)) {
			    if ((rangeFirst % spanDig) == 0) {
				check = checkPass((rangeFirst - diff), ltm, ntm, top, secs);
				if (check) {
				    switch(top) {
					case 59:
					    lastMin = (saved - diff);
					    break;
					case 23:
					    lastHour = (saved - diff);
					    break;
					case 31:
					    lastMDay = (saved - diff);
					    break;
					case 12:
					    lastMonth = (saved - diff);
					    break;
					case 7:
					    lastWDay = (saved - diff);
					    break;
				    }
				    return (rangeFirst - diff);
				}
				break;
			    }
			    rangeFirst++;
			}
			return (saved - diff);
		    }
		}
		rangeFirst++;
	    }
	} else {
	    returnSpans(minutes, words, 121);
	    rangeFirst = strtol(words[0], NULL, 10);
	    spanDig = strtol(words[1], NULL, 10);
	    k = 0;
	    diff = (top == 12) ? 1 : 0;
	    
	    if (isalpha(*words[0])) {
		rangeFirst = checkDayOrMonth(words[0]);
	    }
	    if (diff) { rangeFirst--; }
	    
	    while ((k <= rangeFirst)) {
		if ((k % spanDig) == 0) {
		    check = checkPass((k + diff), ltm, ntm, top, secs);
		    if (check) {
			int saved = k;
			k++;
			while ((k <= rangeFirst)) {
			    if ((k % spanDig) == 0) {
				check = checkPass((k + diff), ltm, ntm, top, secs);
				if (check) {
				    switch(top) {
					case 59:
					    lastMin = (saved + diff);
					    break;
					case 23:
					    lastHour = (saved + diff);
					    break;
					case 31:
					    lastMDay = (saved + diff);
					    break;
					case 12:
					    lastMonth = (saved + diff);
					    break;
					case 7:
					    lastWDay = (saved + diff);
					    break;
				    }
				    return (k + diff);
				}
				break;
			    }
			    k++;
			}
			return (saved + diff);
		    }
		}
		k++;
	    }
	}
    }
    
    return -1000;
}

int finalTimeTest(struct tm *ntm, struct tm *ltm, int thisMin, int thisHour, int thisMday, int thisWday, int thisMonth, int lessDays)
{
    struct tm testTm;
    time_t lastTime, testTime, nowTime;
#ifdef USE_LOG
    FILE *log;
#endif
    
    testTm.tm_sec = 0;
    testTm.tm_min = thisMin;
    testTm.tm_hour = thisHour;
    testTm.tm_mday = (thisMday - lessDays);
    testTm.tm_mon = thisMonth;
    testTm.tm_year = ltm->tm_year;
    testTm.tm_wday = thisWday;
    testTm.tm_isdst = -1;
    
    lastTime = mktime(ltm);
    testTime = mktime(&testTm);
    nowTime = mktime(ntm);
	
#ifdef USE_LOG
    log = fopen(loopLogDef, "a");
    fprintf(log, "Final Time Test\n");
    fprintf(log, "\tlastTime: %s\t\t(%lu)\n", ctime(&lastTime), lastTime);
    fprintf(log, "\ttestTime: %s\t\t(%lu)\n", ctime(&testTime), testTime);
    fprintf(log, "\tnowTime: %s\t\t(%lu)\n", ctime(&nowTime), nowTime);
    fprintf(log, "(%lu >= %lu) && (%lu <= %lu) [%d]\n", testTime, lastTime, testTime, nowTime, ((testTime >= lastTime) && (testTime <= nowTime)));
    fprintf(log, "\n");
    fclose(log);
#endif
    if ((testTime >= lastTime) && (testTime <= nowTime)) {
	return 1;
    }
    return 0;
}

int shouldRun(struct crontab *thisTab, struct tm *ltm, struct tm *ntm, long secs)
{
#ifdef USE_LOG
    int passMin = 0, passHour = 0, passMday = 0, passMonth = 0, passWday = 0;
    FILE *log = fopen(loopLogDef, "a");
#endif
    int thisMin, thisHour, thisMday, thisMonth, thisWday;
    char cronMin[131], cronHour[59], cronMday[73], cronMonth[35], cronWday[25];
#ifndef USE_LOG
    int passMday = 0, passWday = 0;
#endif

    if (secs >= YearDef) {
    /* If, for some insane reason, someone hasn't booted up or left the computer
       asleep for a year, anything definitely passes. Note that I'm generalizing
       a year here - 365 days, when it may actually be 366 for a leap year.
    */
#ifdef USE_LOG
	fprintf(log, "A year or longer: everything passes!\n");
	fclose(log);
#endif
	return 1;
    }

    lastMin = -1; lastHour = -1; lastMDay = -1; lastWDay = -1; lastMonth = 0;
    
    strcpy(cronMin, thisTab->cron_min);
    strcpy(cronHour, thisTab->cron_hour);
    strcpy(cronMday, thisTab->cron_mday);
    strcpy(cronMonth, thisTab->cron_month);
    strcpy(cronWday, thisTab->cron_wday);
    
    thisMin = strtol(cronMin, NULL, 10);
    thisHour = strtol(cronHour, NULL, 10);
    thisMday = strtol(cronMday, NULL, 10);
    thisMonth = strtol(cronMonth, NULL, 10);
    thisWday = strtol(cronWday, NULL, 10);
    
    if (thisWday == 7) {
	thisWday = 0;
    }
    if (isalpha(*cronWday)) {
	thisWday = checkDayOrMonth(cronWday);
    }
    if (isalpha(*cronMonth)) {
	thisMonth = checkDayOrMonth(cronMonth);
    }
    
#ifdef USE_LOG
    fprintf(log, "Testing cron job: (%s)\n", thisTab->cron_command);
    fprintf(log, "Variables:\n\tcron_hour: %s\tcron_min: %s\n\tcron_mday: %s\tcron_month: %s\n\tcron_wday: %s\n\n", cronHour, cronMin, cronMday, cronMonth, cronWday);
    fclose(log);
#endif
    
    if (ListSpanRangeTestDef(cronMin)) {
	thisMin = checkSpan(cronMin, ltm, ntm, 59, secs);
#ifdef USE_LOG
	passMin = (thisMin >= 0);
#endif
#ifndef USE_LOG
	if (thisMin < 0) {
	    return 0;
	}
#endif
    } else if (CheckStarDef(cronMin) || checkPass(thisMin, ltm, ntm, 59, secs)) {
#ifdef USE_LOG
	passMin = 1;
#endif
	if (CheckStarDef(cronMin)) {
	    thisMin = ntm->tm_min;
	}
    }
#ifndef USE_LOG
    else {
	return 0;
    }
#endif
    if (ListSpanRangeTestDef(cronHour)) {
	thisHour = checkSpan(cronHour, ltm, ntm, 23, secs);
#ifdef USE_LOG
	passHour = (thisHour >= 0);
#endif
#ifndef USE_LOG
	if (thisHour < 0) {
	    return 0;
	}
#endif
    } else if (CheckStarDef(cronHour) || checkPass(thisHour, ltm, ntm, 23, secs)) {
#ifdef USE_LOG
	passHour=1;
#endif
	if (CheckStarDef(cronHour)) {
	    thisHour = ntm->tm_hour;
	}
    }
#ifndef USE_LOG
    else {
	return 0;
    }
#endif
    if (ListSpanRangeTestDef(cronMday)) {
	thisMday = checkSpan(cronMday, ltm, ntm, 31, secs);
	passMday = (thisMday >= 0);
    } else if (CheckStarDef(cronMday) || checkPass(thisMday, ltm, ntm, 31, secs)) {
	passMday = 1;
	if (CheckStarDef(cronMday)) {
	    thisMday = ntm->tm_mday;
	}
    }
    if (ListSpanRangeTestDef(cronMonth)) {
	thisMonth = checkSpan(cronMonth, ltm, ntm, 12, secs);
#ifdef USE_LOG
	passMonth = (thisMonth >= 0);
#endif
#ifndef USE_LOG
	if (thisMonth < 0) {
	    return 0;
	}
#endif
    } else if (CheckStarDef(cronMonth) || checkPass(thisMonth, ltm, ntm, 12, secs)) {
#ifdef USE_LOG
	passMonth = 1;
#endif
	if (CheckStarDef(cronMonth)) {
	    thisMonth = ntm->tm_mon;
	}
    }
#ifndef USE_LOG
    else {
	return 0;
    }
#endif
    if (ListSpanRangeTestDef(cronWday)) {
	thisWday = checkSpan(cronWday, ltm, ntm, 7, secs);
	passWday = (thisWday >= 0);
	if (!passMday && passWday && !CheckEmptyDef(cronMday)) {
#ifdef USE_LOG
	    log = fopen(loopLogDef, "a");
	    fprintf(log, "Setting passMday to pass.\n");
	    fclose(log);
	    passMday = 1;
#endif
	    thisMday = ntm->tm_mday;
	}
#ifndef USE_LOG
	else {
	    return 0;
	}
#endif
    } else if (CheckStarDef(cronWday) || checkPass(thisWday, ltm, ntm, 7, secs)) {
	passWday = 1;
	if (CheckStarDef(cronWday)) {
	    thisWday = ntm->tm_wday;
	}
	if (!passMday && passWday && !CheckEmptyDef(cronWday) && !CheckEmptyDef(cronMday)) {
#ifdef USE_LOG
	    log = fopen(loopLogDef, "a");
	    fprintf(log, "Setting passMday to pass.\n");
	    fclose(log);
	    passMday = 1;
#endif
	    thisMday = ntm->tm_mday;
	}
    } else if (passMday && !CheckEmptyDef(cronMday) && !CheckEmptyDef(cronWday)) {
#ifdef USE_LOG
	log = fopen(loopLogDef, "a");
	fprintf(log, "Setting passWday to pass.\n");
	fclose(log);
	passWday = 1;
#endif
	thisWday = ntm->tm_wday;
    }
#ifndef USE_LOG
    else {
	return 0;
    }
#endif
    if (!CheckEmptyDef(cronMonth)) {
	thisMonth--;
    }
    lastMonth--;
    
#ifdef USE_LOG
    log = fopen(loopLogDef, "a");
    fprintf(log, "passMin: %d passHour: %d passMday: %d passMonth: %d passWday: %d\n", passMin, passHour, passMday, passMonth, passWday);
    fprintf(log,"\n");
    fclose(log);
    
    if (passMin && passHour && passMday && passMonth && passWday) {
#else
    {
#endif
	/*Thus far, it thinks everything has passed.
	  Now I'm going to *really* test that. */
	int dayDiff = (ntm->tm_yday - ltm->tm_yday);
	int weekDiff = (ntm->tm_wday - thisWday);
	int noMday = (CheckEmptyDef(cronMday));
	int noWday = (CheckEmptyDef(cronWday));
	int timePassed = 0;
	
	if (dayDiff < 0) {
	    dayDiff = ((ntm->tm_yday + 365) - ltm->tm_yday);
	}
	if (weekDiff < 0) {
	    weekDiff = ((ntm->tm_wday + 6) - thisWday);
	}
	
	if (ltm->tm_year != ntm->tm_year && dayDiff > 62) {
	    thisMonth = (ltm->tm_mon + 1);
	}

	if (noMday && noWday) {
	    int n;
	    for (n = 0; n <= dayDiff; n++) {
		timePassed = finalTimeTest(ntm, ltm, thisMin, thisHour, thisMday, thisWday, thisMonth, n);
		if (timePassed) {
		    return 1;
		}
	    }
	    /* Sigh.  Fix one bug, make 20 worse ones.  I found out I needed to check a
	       day ahead when there are spans, for it could miss out on running *this* day.
	       But doing that means that now I have to go back and check to see if the first time it
	       passed should have actually passed, as well.  Pain in the butt. */
	    if ((lastMin != -1) || (lastHour != -1) || (lastMDay != -1) || (lastWDay != -1) || (lastMonth != -1)) {
		for (n = 0; n <= dayDiff; n++) {
		    timePassed = finalTimeTest(ntm, ltm, (lastMin != -1) ? lastMin : thisMin, (lastHour != -1) ? lastHour : thisHour, (lastMDay != -1) ? lastMDay : thisMday, (lastWDay != -1) ? lastWDay : thisWday, (lastMonth != -1) ? lastMonth : thisMonth, n);
		    if (timePassed) {
			return 1;
		    }
		}
	    }
	} else if (noMday && !noWday) {
	    timePassed = finalTimeTest(ntm, ltm, thisMin, thisHour, thisMday, thisWday, thisMonth, weekDiff);
	    if (timePassed) {
		return 1;
	    }
	    if ((lastMin != -1) || (lastHour != -1) || (lastMDay != -1) || (lastWDay != -1) || (lastMonth != -1)) {
		if (lastWDay != -1) {
		    weekDiff = (ntm->tm_wday - lastWDay);
		    if (weekDiff < 0) {
			weekDiff = ((ntm->tm_wday + 6) - lastWDay);
		    }
		}
		timePassed = finalTimeTest(ntm, ltm, (lastMin != -1) ? lastMin : thisMin, (lastHour != -1) ? lastHour : thisHour, (lastMDay != -1) ? lastMDay : thisMday, (lastWDay != -1) ? lastWDay : thisWday, (lastMonth != -1) ? lastMonth : thisMonth, weekDiff);
		if (timePassed) {
		    return 1;
		}
	    }
	} else if (!noMday && noWday) {
	    timePassed = finalTimeTest(ntm, ltm, thisMin, thisHour, thisMday, thisWday, thisMonth, 0);
	    if (timePassed) {
		return 1;
	    }
	    if ((lastMin != -1) || (lastHour != -1) || (lastMDay != -1) || (lastWDay != -1) || (lastMonth != -1)) {
		timePassed = finalTimeTest(ntm, ltm, (lastMin != -1) ? lastMin : thisMin, (lastHour != -1) ? lastHour : thisHour, (lastMDay != -1) ? lastMDay : thisMday, (lastWDay != -1) ? lastWDay : thisWday, (lastMonth != -1) ? lastMonth : thisMonth, 0);
		if (timePassed) {
		    return 1;
		}
	    }
	} else {
	    /* !noMday && !noWday */
	    /* If it has a month day and a week day set,
	       either can pass, so we need to test for both cases. */
	    timePassed = finalTimeTest(ntm, ltm, thisMin, thisHour, thisMday, thisWday, thisMonth, 0);
	    if (timePassed) {
		return 1;
	    }
	    timePassed = finalTimeTest(ntm, ltm, thisMin, thisHour, thisMday, thisWday, thisMonth, weekDiff);
	    if (timePassed) {
		return 1;
	    }
	    if ((lastMin != -1) || (lastHour != -1) || (lastMDay != -1) || (lastWDay != -1) || (lastMonth != -1)) {
		if (lastWDay != -1) {
		    weekDiff = (ntm->tm_wday - lastWDay);
		    if (weekDiff < 0) {
			weekDiff = ((ntm->tm_wday + 6) - lastWDay);
		    }
		}
		timePassed = finalTimeTest(ntm, ltm, (lastMin != -1) ? lastMin : thisMin, (lastHour != -1) ? lastHour : thisHour, (lastMDay != -1) ? lastMDay : thisMday, (lastWDay != -1) ? lastWDay : thisWday, (lastMonth != -1) ? lastMonth : thisMonth, 0);
		if (timePassed) {
		    return 1;
		}
		timePassed = finalTimeTest(ntm, ltm, (lastMin != -1) ? lastMin : thisMin, (lastHour != -1) ? lastHour : thisHour, (lastMDay != -1) ? lastMDay : thisMday, (lastWDay != -1) ? lastWDay : thisWday, (lastMonth != -1) ? lastMonth : thisMonth, weekDiff);
		if (timePassed) {
		    return 1;
		}
	    }
	}
    }

    return 0;
}
