/*
    loop.c
	entry point to timeLoop
    
    Copyright 2003 - 2004 Night Productions, by Michael A. Nickerson (aka Darkshadow).
    All rights reserved.
    Licensed under the Open Software License version 2.0
*/


#include <stddef.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <sys/wait.h>
#include <sys/signal.h>
#include <fcntl.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <paths.h>
#include "loop.h"

#ifdef HAVE_SYSLOG_H
    #include <syslog.h>
#else
    #define LOG_CRIT 0
    #define LOG_INFO 0
    #define LOG_ERR 0
    #define LOG_PID 0
    #define LOG_DAEMON 0
#endif /* defined HAVE_SYSLOG_H */

#ifdef HAVE_FRAMEWORKS
    #include <IOKit/IOKitLib.h>
    #include <IOKit/pwr_mgt/IOPMLib.h>
    #include <IOKit/pwr_mgt/IOPM.h>
    #include <IOKit/IOReturn.h>
    #include <CoreFoundation/CFArray.h>
    #include <CoreFoundation/CFBase.h>
    #include <CoreFoundation/CFNumber.h>
    #include <CoreFoundation/CFData.h>
    #include <mach/mach_init.h>
#endif /* defined HAVE_FRAMEWORKS */

#define READ_END 0
#define WRITE_END 1
#define STDIN 0
#define STDOUT 1
#define STDERR 2

char *ProgName = NULL;

void    releaseTabs __P((struct crontab *)),
	loadSysTab __P((struct crontab *)),
	loadUserTab __P((struct crontab *)),
	loadDatabase __P((struct cronrep *)),
	reloadDatabase __P((struct cronrep *, int));
int     shouldRun __P((struct crontab *, struct tm *, struct tm *, long));
char *  envget __P((char *env, char *environment[]));

struct cronrep mainTab;

void sighup_handler __P((int)),
     sigterm_handler __P((int));
     
void child(struct crontab *runTab)
{
    int in[2], out[2];
    int children = 0;
    char *jobData = NULL, *jPtr = NULL;
    
    pipe(in);
    pipe(out);
    
    signal(SIGCHLD, SIG_DFL);
    
    for (jPtr = runTab->cron_command; *jPtr != '\0'; *jPtr++) {
	if (*jPtr == '%') {
	    *jPtr--;
	    if (*jPtr != '\\') {
		*jPtr++;
		*jPtr='\0';
		*jPtr++;
		jobData = jPtr;
		break;
	    } else {
		*jPtr++;
	    }
	}
    }
    
    switch(vfork()) {
	case -1:
	    timeLog(LOG_CRIT, "Couldn't fork process!");
	    _exit(1);
	    break;
	case 0:
	    {
		char *shell = envget("SHELL", runTab->environment);
		char *home = envget("HOME", runTab->environment);
		
		close(in[WRITE_END]);
		close(out[READ_END]);
		close(STDIN);
		dup2(in[READ_END], STDIN);
		close(STDOUT);
		dup2(out[WRITE_END], STDOUT);
		close(STDERR);
		dup2(STDOUT, STDERR);
		
		close(in[READ_END]);
		close(out[WRITE_END]);
		
		setsid();
		setgid(runTab->cron_user_gid);
		setlogin(runTab->cron_user);
		setuid(runTab->cron_user_uid);
		chdir(home);
		
		timeLog(LOG_INFO, "(%s): Running missed cron task (%s)",runTab->cron_user, runTab->cron_command);
		timeCloseLog();
		execle(shell, shell, "-c", runTab->cron_command, NULL, runTab->environment);
		timeLog(LOG_ERR, "Could not run command (%s)", runTab->cron_command);
		_exit(1);
	    }
	    break;
	default:
	    {
		FILE *input;
		int inputChar;
		children++;
		
		close(in[READ_END]);
		close(out[WRITE_END]);
		
		if (jobData) {
		    switch (fork()) {
			case -1:
			    timeLog(LOG_ERR, "Could not send input to cron job");
			    _exit(1);
			case 0:
			{
			    FILE *outPipe = fdopen(in[WRITE_END], "w");
			    close(out[READ_END]);
			    
			    if (outPipe == NULL) {
				timeLog(LOG_ERR, "Could not send input to cron job");
				_exit(1);
			    }
			    
			    for (jPtr = jobData; *jPtr != '\0'; *jPtr++) {
				if (*jPtr == '\\') {
				    *jPtr++;
				    if (*jPtr == '%') {
					putc(*jPtr, outPipe);
					continue;
				    } else {
					putc('\\', outPipe);
					putc(*jPtr, outPipe);
					continue;
				    }
				} else if (*jPtr == '%') {
				    putc('\n', outPipe);
				} else {
				    putc(*jPtr, outPipe);
				}
			    }
			    
			    fclose(outPipe);
			    close(in[WRITE_END]);
			    _exit(0);
			}
			default:
			    children++;
			    break;
		    }
		}
		
		close(in[WRITE_END]);
		
		input = fdopen(out[READ_END], "r");
		
		if (input == NULL) {
		    timeLog(LOG_ERR, "Could not read task's output");
		    close(out[READ_END]);
		    break;
		}
		
		inputChar=getc(input);
		
		if (inputChar != EOF) {
		    FILE *mail = NULL;
		    char *mailUser = envget("MAILTO", runTab->environment);
		    int mailPipe[2];
		    
		    if (pipe(mailPipe) < 0) {
			timeLog(LOG_ERR, "Sending mail failed.");
			fclose(input);
			close(out[READ_END]);
			break;
		    }
		    
		    if (mailUser != NULL) {
			if (!*mailUser) {
			    mailUser = NULL;
			}
		    } else {
			mailUser = runTab->cron_user;
		    }
		    
		    if (mailUser) {
			switch(fork()) {
			    case -1:
				close(mailPipe[READ_END]);
				close(mailPipe[WRITE_END]);
				_exit(1);
			    case 0:
			    {
				char *home = envget("HOME", runTab->environment);
				if (mailPipe[READ_END] != STDIN) {
				    close(STDIN);
				    dup2(mailPipe[READ_END], STDIN);
				    close(mailPipe[READ_END]);
				}
				/* Anyone know why I *must* close STDOUT and STDERR
				 * here?  Without doing the below, sending mail
				 * fails...
				 */
				close(STDOUT);
				open(_PATH_DEVNULL, O_RDWR);
				close(STDERR);
				open(_PATH_DEVNULL, O_RDWR);
				close(mailPipe[WRITE_END]);
				
				setsid();
				setgid(runTab->cron_user_gid);
				setlogin(runTab->cron_user);
				setuid(runTab->cron_user_uid);
				chdir(home);
				
				timeCloseLog();
		    
				execl(mailProgDef, mailProgDef, "-odi", "-oem", "-oi", "-t", NULL);
				_exit(1);
			    }
			    default:
				children++;
				break;
			}
		    
			mail=fdopen(mailPipe[WRITE_END], "w");
			close(mailPipe[READ_END]);
		    
			if (mail == NULL) {
			    timeLog(LOG_ERR, "Sending mail failed.");
			    close(mailPipe[WRITE_END]);
			    break;
			}
			
			{
			    char hostname[256];
			    int x = 0;
			    
			    gethostname(hostname, 256);
			    fprintf(mail, "From: %s (timeLoop Daemon)\n", runTab->cron_user);
			    fprintf(mail, "To: %s\n",mailUser);
			    fprintf(mail, "Subject: %s - timeLoop <%s@%s>\n", runTab->cron_command, runTab->cron_user, hostname);
			    while ((runTab->environment[x] != NULL)) {
				fprintf(mail, "X-Env: <%s>\n", runTab->environment[x]);
				x++;
			    }
			    fprintf(mail, "\n");
			    putc(inputChar, mail);
			}
		    }
		    
		    while ((inputChar = getc(input)) != EOF) {
			if (mailUser) {
			    putc(inputChar, mail);
			}
		    }
		    if (mailUser) {
			fprintf(mail, "\n");
			fflush(mail);
			fclose(mail);
		    }
		    close(mailPipe[WRITE_END]);
		}
		fclose(input);
		close(out[READ_END]);
	    }
	    for (; children > 0; children--) {
		int waiter;
		wait(&waiter);
	    }
	    break;
    }
}

void run(struct crontab *runTab)
{
#ifdef HAVE_FRAMEWORKS
    if (runTab->batteryFlags) {
	CFArrayRef array;
	mach_port_t master = 0;
	io_connect_t ioconn = 0;
	
	IOMasterPort(bootstrap_port, &master);
	ioconn=IOPMFindPowerManagement(master);
	
	if (IOPMCopyBatteryInfo(master, &array) == kIOReturnSuccess) {
	    CFDictionaryRef dict;
	    CFNumberRef num;
	    int batteryFlag;
	    
	    dict = CFArrayGetValueAtIndex(array, 0);
	    num = CFDictionaryGetValue(dict, CFSTR(kIOBatteryFlagsKey));
	    CFNumberGetValue(num, kCFNumberLongType, &batteryFlag);
	    
	    if (!(batteryFlag & kIOBatteryChargerConnect)) {
		timeLog(LOG_INFO, "Not running job (%s) while on battery power.", runTab->cron_command);
		return;
	    }
	}
    }
#endif
    switch(fork()) {
	case -1:
	    timeLog(LOG_ERR, "Couldn't run job (%s): internal error.", runTab->cron_command);
	    break;
	case 0:
	    child(runTab);
	    _exit(0);
	    break;
	default:
	    break;
    }
    
}

void time_loop()
{
    time_t now;
    time_t lastTime=time(NULL);

    while(1) {
	now=time(NULL);
	if ((now - lastTime) > 60) {
	    struct stat sysSt, userSt;
	    struct tm ltime, ntime;
	    struct crontab *thisTab;
	    FILE *spool=fopen(loopSpoolFileDef, "w");
	    
#ifdef USE_LOG
	    FILE *log;
#endif
	    fprintf(spool, "%lu\n", (unsigned long)now);
	    fclose(spool);
	    
	    localtime_r(&now, &ntime);
	    localtime_r(&lastTime, &ltime);
	    stat(systemCrontabFileDef, &sysSt);
	    stat(userCrontabDirectoryDef, &userSt);
	    
	    if(mainTab.sysTabModDate != sysSt.st_mtime) {
		reloadDatabase(&mainTab, SystemTabsDef);
		mainTab.sysTabModDate = sysSt.st_mtime;
	    }
	    if(mainTab.userTabModDate != userSt.st_mtime) {
		reloadDatabase(&mainTab, UserTabsDef);
		mainTab.userTabModDate = userSt.st_mtime;
	    }
	    
#ifdef USE_LOG
	    log = fopen(loopLogDef, "w");
	    fprintf(log, "timeLoop testing at: %s", ctime(&now));
	    fprintf(log, "last ran at: %s\n", ctime(&lastTime));
	    fprintf(log, "Testing system crontab:\n\n");
	    fclose(log);
#endif
	    
	    thisTab=mainTab.systemTab;
	    
	    while(thisTab->nextTab != NULL) {
		if (shouldRun(thisTab, &ltime, &ntime, (now - lastTime))) {
		    run(thisTab);
		}
		thisTab = thisTab->nextTab;
	    }
#ifdef USE_LOG
	    log = fopen(loopLogDef, "a");
	    fprintf(log, "Testing user crontabs:\n\n");
	    fclose(log);
#endif
	    
	    thisTab = mainTab.userTab;
	    
	    while(thisTab->nextTab != NULL) {
#ifdef USE_LOG
		log = fopen(loopLogDef,"a");
		fprintf(log, "User: %s\n", thisTab->cron_user);
		fclose(log);
#endif
		if (shouldRun(thisTab, &ltime, &ntime, (now - lastTime))) {
		    run(thisTab);
		}
		thisTab = thisTab->nextTab;
	    }
	}
	lastTime = now;
	sleep(60);
    }
}

void firstStart()
{
    unsigned long tme;
    time_t lastTime;
    time_t now;
    struct tm ltm;
    struct tm ntm;
    struct crontab *thisTab = NULL;
    FILE *spool = fopen(loopSpoolFileDef, "r");
    
#ifdef USE_LOG
    FILE *log;
#endif
    
    if (spool == NULL) {
	/*very first run - spool doesn't exist yet*/
	now = time(NULL);
	lastTime = time(NULL);
	mkdir(loopSpoolDirDef, 0755);
    } else {
	fscanf(spool, "%lu", &tme);
	fclose(spool);
	lastTime = (time_t)tme;
	now = time(NULL);
    }
    spool = fopen(loopSpoolFileDef, "w");
    fprintf(spool, "%lu\n", (unsigned long)now);
    fclose(spool);
    
#ifdef USE_LOG
    log = fopen(loopLogDef, "w");
    fprintf(log, "timeLoop started at: %s", ctime(&now));
    fprintf(log, "last ran at: %s\n", ctime(&lastTime));
    fprintf(log, "Testing system crontab:\n\n");
    fclose(log);
#endif
    
    localtime_r(&now, &ntm);
    localtime_r(&lastTime, &ltm);
    
    thisTab = mainTab.systemTab;
	    
    while(thisTab->nextTab != NULL) {
	if (shouldRun(thisTab, &ltm, &ntm, (now - lastTime))) {
	    run(thisTab);
	}
	thisTab = thisTab->nextTab;
    }
    
#ifdef USE_LOG
    log = fopen(loopLogDef, "a");
    fprintf(log, "Testing user crontabs:\n\n");
    fclose(log);
#endif
	    
    thisTab = mainTab.userTab;
    
    while(thisTab->nextTab != NULL) {
#ifdef USE_LOG
	log = fopen(loopLogDef, "a");
	fprintf(log, "User: %s\n", thisTab->cron_user);
	fclose(log);
#endif
	if (shouldRun(thisTab, &ltm, &ntm, (now - lastTime))) {
	    run(thisTab);
	}
	thisTab = thisTab->nextTab;
    }
}
    

int main (int argc, char *argv[])
{
    time_t now;
    FILE *pidFile = NULL;
    int pidRef = 0;
    if (getuid() != 0) {
	printf("Must be run with root privileges.\n");
	return 1;
    }

    ProgName = argv[0];
    
    now = time(NULL);
    
    daemon(0,0);

    timeOpenLog(ProgName, LOG_PID, LOG_DAEMON);
    timeLog(LOG_INFO, "Starting at %s", ctime(&now));
    
    if ((pidRef = open(loopPidFileDef, O_RDWR | O_CREAT | O_EXCL, 0644)) < 0) {
	int pidNumber;
	pidFile = fopen(loopPidFileDef, "r");
	fscanf(pidFile, "%d", &pidNumber);
	fclose(pidFile);
	timeLog(LOG_INFO, "Unable to create pid file, another instance of %s may be running with a PID of %d.  Exiting...", ProgName, pidNumber);
	exit(1);
    }
    pidFile = fdopen(pidRef, "w+");
    fprintf(pidFile, "%d\n", getpid());
    fflush(pidFile);
    fclose(pidFile);
    
    signal(SIGHUP, sighup_handler);
    signal(SIGTERM, sigterm_handler);
    signal(SIGCHLD, SIG_IGN);
    
    loadDatabase(&mainTab);
    
    firstStart();
    
    time_loop();
    
    return 0;
}

void sighup_handler(int sig)
{
    reloadDatabase(&mainTab, SystemTabsDef);
    reloadDatabase(&mainTab, UserTabsDef);
}

void sigterm_handler(int sig)
{
    FILE *spool = fopen(loopSpoolFileDef, "w");
    time_t now = time(NULL);
    fprintf(spool, "%lu\n", (unsigned long)now);
    fclose(spool);
    unlink(loopPidFileDef);
    timeLog(LOG_INFO, "Stopping at %s", ctime(&now));
    timeCloseLog();
    exit(0);
}

extern void timeLog(int priority, const char *message, ...)
{
    va_list args;
#ifdef HAVE_SYSLOG_H
    va_start(args, message);
    vsyslog(priority, message, args);
    va_end(args);
#else
    FILE *timeLogFile = fopen(timeLogFileDef, "a");
    char theTime[27];
    time_t nowTime = time(NULL);
    
    ctime_r(&nowTime, (char *)&theTime);
    theTime[24]='\0';
    fprintf(timeLogFile, "%s: ", theTime);
    va_start(args, message);
    vfprintf(timeLogFile, message, args);
    va_end(args);
    fclose(timeLogFile);
#endif
}

extern void timeOpenLog(const char *ident, int logopt, int facility)
{
#ifdef HAVE_SYSLOG_H
    openlog(ident, logopt, facility);
#endif
}

extern void timeCloseLog(void)
{
#ifdef HAVE_SYSLOG_H
    closelog();
#endif
}